#include "configparser.h"
#include "libfiles.h"
#include "libproc.h"
#include "libtime.h"

#include <string>
#include <iostream>
#include <cstdlib>
#include <sys/stat.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <cstdio>
#include <time.h>

using namespace std;

ConfigParser cfg;

int idlenesslimit;
int timelimit;
int movestodointl;
int memorylimit;
string p1exec;
string p2exec;
string cexec;
string p1cmdline;
string p2cmdline;
string ccmdline;

long long statebegtime;

void LoadConfig()
{
	try{
		cfg.Load("game_run.cfg");
		idlenesslimit = atoi(cfg.GetParam("IdlenessLimit").c_str());
		timelimit = atoi(cfg.GetParam("TimeLimit").c_str());
		movestodointl = atoi(cfg.GetParam("MovesToDoInTL").c_str());
		memorylimit = atoi(cfg.GetParam("MemoryLimit").c_str());
		p1exec = cfg.GetParam("Player1");
		p2exec = cfg.GetParam("Player2");
		cexec = cfg.GetParam("Checker");
		p1cmdline = cfg.GetParam("Player1CommandLine");
		p2cmdline = cfg.GetParam("Player2CommandLine");
		ccmdline = cfg.GetParam("CheckerCommandLine");
	}
	catch(string s)
	{
		cerr << "Error during config load: " << s << "\n";
	}
}

enum States {
	READP1 = 0,
	READCHECKP1 = 1,
	READP2 = 2,
	READCHECKP2 = 3
};

// descriptors for reading/writing
int p1read, p1write;
int p2read, p2write;
int cread, cwrite;
// process IDs
int p1id, p2id, cid;
bool p1dead, p2dead, cdead;
int p1code, p2code, ccode;

// SIGCHLD handler
void OnSIGCHLD(int signum)
{
    int pid, status, olderr;
    olderr = errno;
    while((pid = waitpid(WAIT_ANY, &status, WNOHANG)) != 0) 
	{
        if (pid == -1) 
		{
            if (errno != ECHILD) 
			{
                exit(1);
            }
            break;
        }
		cerr << "Process " << pid << " dead now with status = " << status << "\n";
		if (pid == p1id) 
		{
			p1dead = true;
			p1code = status;
		}
		if (pid == p2id) 
		{
			p2dead = true;
			p2code = status;
		}
		if (pid == cid)
		{
			cdead = true;
			ccode = status;
		}
    }
    errno = olderr;
}

// carefully exits (kills all child processes)
void Exit()
{
	if (!p1dead) kill(p1id, SIGKILL);
	if (!p2dead) kill(p2id, SIGKILL);
	if (!cdead) kill(cid, SIGKILL);
	exit(0);
}

string alreadyread;
int cstate;
bool firstmove;
int p1moves, p2moves;

void CheckIdlenessLimit()
{
	double cs = (GetCurrentTime() - statebegtime) / (double)1000000000.0;
	double il = (idlenesslimit + 0.0) / 1000.0;
	if (cs > il)
	{
		if (cstate == READP1)
		{
			cout << "Second Wins\nPlayer 1 idleness limit exceeded\n";
			Exit();
		}
		else if (cstate == READP2)
		{
			cout << "First Wins\nPlayer 2 idleness limit exceeded\n";
			Exit();
		}
		else
		{
			cout << "Fail\nChecker idleness limit exceeded\n";
			Exit();
		}
	}
}

void CheckTimeLimit()
{
	int p1cycles = (int)(GetProcessTime(p1id) / (timelimit / 1000.0) + 1e-8);
	if (p1cycles * movestodointl > p1moves)
	{
		cout << "Second Wins\nPlayer 1 time limit exceeded\n";
		Exit();
	}
	int p2cycles = (int)(GetProcessTime(p2id) / (timelimit / 1000.0) + 1e-8);
	if (p2cycles * movestodointl > p2moves)
	{
		cout << "First Wins\nPlayer 2 time limit exceeded\n";
		Exit();
	}
}

void CheckMemoryLimit()
{
	long long p1m = GetProcessMemory(p1id);
	if (p1m > memorylimit)
	{
		cout << "Second Wins\nPlayer 1 memory limit exceeded\n";
		Exit();
	}
	long long p2m = GetProcessMemory(p2id);
	if (p2m > memorylimit)
	{
		cout << "First Wins\nPlayer 2 memory limit exceeded\n";
		Exit();
	}
}

// should check timelimits, memorylimits and so on...
void CheckEverything()
{
	// check dead processes
	if (p1dead)
	{
		cout << "Second Wins\n";
		cout << "Player 1 dead\n";
		Exit();
	}
	if (p2dead)
	{
		cout << "First Wins\n";
		cout << "Player 2 dead\n";
		Exit();
	}
	if (cdead)
	{
		if (ccode != 0)
		{
			if (cstate == READCHECKP1)
			{
				cout << "Second Wins\n";
				cout << "Checker exited with code " << ccode << " after first player move\n";
				Exit();
			}
			else if (cstate == READCHECKP2)
			{
				cout << "First Wins\n";
				cout << "Checker exited with code " << ccode << " after second player move\n";
				Exit();
			}
		}
		else 
		{
			return;
		}
	}
	CheckIdlenessLimit();
	CheckTimeLimit();
	CheckMemoryLimit();
}

string ParseCheckerOutput()
{
	if (firstmove) return alreadyread;
	// find first 10
	int i, j;
	string verd = "";
	string toplayer;
	for (i = 0; i < alreadyread.length(); i++)
	{
		if (alreadyread[i] == 10)
		{
			for (j = 0; j < i; j++) 
			{
				if (alreadyread[j] != 13) verd += alreadyread[j];
			}
			toplayer = alreadyread.substr(i + 1);
			break;
		}
	}
	if (verd == "Move OK")
	{
		return toplayer;
	}
	cout << verd << "\n" << toplayer << "\n";
	CopyFile("./Temp/check/game.log", "game.log");
	Exit();
}

bool ReadFromCurrent()
{
	int cfh = 0;
	if (cstate == READP1) cfh = p1read;
	else if (cstate == READP2) cfh = p2read;
	else cfh = cread;
	char buf[1025];
	int res = read(cfh, buf, 1024);
	while (res > 0)
	{
		buf[res] = 0;
		int i;
		for (i = 0; i < res; i++)
		{
			if (buf[i] != 13) alreadyread += buf[i];
		}
		CheckEverything();
		res = read(cfh, buf, 1024);
	}
	CheckEverything();
	if (alreadyread.length() > 2)
	{
		if (alreadyread[alreadyread.length() - 1] == 10 &&
			alreadyread[alreadyread.length() - 2] == 10) return true;
	}
	return false;
}

void WriteToProcess(int wto)
{
	int pos = 0;
	while (pos < alreadyread.length())
	{
		int res = write(wto, &(alreadyread.c_str()[pos]), alreadyread.length() - pos);
		if (res < 0)
		{
			if (errno != EAGAIN)
			{
				CheckEverything();
				throw (string)("Some strange thing during writing");
			}
			usleep(50 * 1000); // 5/100 of a second
		}
		else pos += res;
		CheckEverything();
	}
}

int main()
{
	struct sigaction sa;
    
    // set SIGCHLD handler
    sigemptyset(&sa.sa_mask);
    sa.sa_handler = OnSIGCHLD;
    sa.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    if (sigaction(SIGCHLD, &sa, NULL) == -1) 
    {
		cerr << "Could not set SIGCHLD handler\n";
        exit(1);
    }
    
	LoadConfig();
	// create needed directories
	mkdir("Temp", S_IRWXU);
	mkdir("Temp/player1", S_IRWXU);
	mkdir("Temp/player2", S_IRWXU);
	mkdir("Temp/check", S_IRWXU);
	CopyFile(p1exec, "Temp/player1/" + p1exec);
	CopyFile(p2exec, "Temp/player2/" + p2exec);
	CopyFile(cexec, "Temp/check/" + cexec);
	// run everybody...
	p1dead = p2dead = cdead = false;
	RunProcess(p1exec + " 1", "Temp/player1", p1id, p1read, p1write);
	RunProcess(p2exec + " 2", "Temp/player2", p2id, p2read, p2write);
	RunProcess(cexec, "Temp/check", cid, cread, cwrite);
	cerr << "Runned player1, id=" << p1id << " player2, id = " << p2id << " check, id = " << cid << "\n";
	// main loop 
	alreadyread = "";
	cstate = READCHECKP2;
	firstmove = true;
	int curmove = 0;
	statebegtime = GetCurrentTime();
	p1moves = p2moves = 0;
	while (1)
	{
		if (ReadFromCurrent())
		{
			//cerr << "reading finished...\n";
			//cerr << "already read = " << alreadyread << "\n";
			// if we finished reading from current process
			int wto = cwrite;
			if (cstate == READP1 || cstate == READP2)
			{
				wto = cwrite;
				curmove++;
				cerr << "It is move #" << curmove << "\n";
				if (cstate == READP1) p1moves++;
				else p2moves++;
			}
			else
			{
				// may be it will exit
				alreadyread = ParseCheckerOutput();
				if (cstate == READCHECKP1) wto = p2write;
				else if (cstate == READCHECKP2) wto = p1write;
			}
			cstate = (cstate + 1) % 4;
			statebegtime = GetCurrentTime();
			WriteToProcess(wto);
			alreadyread = "";
			firstmove = false;
		}
		CheckEverything();
		usleep(50 * 1000); // 5/100 of a second
	}
	Exit();
	return 0;
}